#include <cmath>
#include <math.h>

#include "camera.h"

#include "render.h"

/* class Xr_Camera */

Xr_Camera::Xr_Camera() : C_projection(new Xr_MATRIX_4X4), C_box(new Xr_FrustumBox), C_fov(70), C_far(1000), C_near(0.001), C_angle({0})
{	
	computeProjection();
	C_dir.setCoords(1, 0, 0);
	C_right.setCoords(0, 0, -1);
	C_left = -C_right;
	C_up.setCoords(0, 1, 0);
	C_down = -C_up;
}

Xr_Camera::~Xr_Camera()
{
	delete C_projection;
	delete C_box;
}

void Xr_Camera::configModelview(Xr_MATRIX_4X4 *modelview)
{
	float aX = M_PI * (C_angle[0] / (float) 180);
	float aY = M_PI * (C_angle[1] / (float) 180);

	C_eye.setCoords(C_near * (cos(aY) * cos(aX)), C_near * sin(aY), C_near * (cos(aY) * sin(aX)));
	C_eye += C_pos;
	
	C_dir = C_pos - C_eye;
	C_right.setCoords(-C_dir.getZ(), 0, C_dir.getX());
	C_left = -C_right;
	C_up = C_right * C_dir;
	C_down = -C_up;
	
	C_dir.normalize();
	C_right.normalize();
	C_left.normalize();
	C_up.normalize();
	C_down.normalize();
	
	C_box->genBox(C_pos, C_dir, C_right, C_left, C_up, C_down);
	modelview->lookAt(C_eye[0], C_eye[1], C_eye[2], C_pos[0], C_pos[1], C_pos[2], 0, 1, 0);
}

void Xr_Camera::computeProjection()
{
	float ratio = (float) Xr_Renderer::getInstance()->getImageWidth() / Xr_Renderer::getInstance()->getWindowHeight();
	C_projection->loadIdentity();
	C_projection->loadProjection(C_fov, ratio, C_near, C_far);

	C_box->setBoxSize(C_near, C_far, C_fov, ratio);
}

void Xr_Camera::setPos(float x, float y, float z)
{
	C_pos.setCoords(x, y, z);
}

void Xr_Camera::relPos(float x, float y, float z)
{
	C_pos += Xr_Vector3D(x, y, z);
}

void Xr_Camera::setAngle(float x, float y)
{
	C_angle[0] = x;

	if(y > 89)
	{
		C_angle[1] = 89;
	}
	else if(y < -89)
	{
		C_angle[1] = -89;
	}
}

void Xr_Camera::relAngle(float x, float y)
{
	C_angle[0] += x;
	C_angle[1] += y;
	
	if(C_angle[1] > 89)
	{
		C_angle[1] = 89;
	}
	else if(C_angle[1] < -89)
	{
		C_angle[1] = -89;
	}
}

void Xr_Camera::setFov(float fov)
{
	C_fov = fov;
	
	if(C_fov > 180)
	{
		C_fov = 180;
	}
	if(C_fov < 1)
	{
		C_fov = 1;
	}
	
	computeProjection();
}

void Xr_Camera::relFov(float fov)
{
	C_fov += fov;
	
	if(C_fov > 180)
	{
		C_fov = 180;
	}
	if(C_fov < 1)
	{
		C_fov = 1;
	}
	
	computeProjection();
}

void Xr_Camera::setFar(float far)
{
	C_far = far;
}

void Xr_Camera::setNear(float near)
{
	C_near = near;
}

void Xr_Camera::moveForward(float n)
{
	C_pos += C_dir * n;
}

void Xr_Camera::moveBackward(float n)
{
	C_pos -= C_dir * n;
}

void Xr_Camera::moveRight(float n)
{
	C_pos += C_right * n;
}

void Xr_Camera::moveLeft(float n)
{
	C_pos += C_left * n;
}

void Xr_Camera::moveUpDown(float n)
{
	relPos(0, n, 0);
}

Xr_Vector3D Xr_Camera::getPos() const
{
	return C_pos;
}

float Xr_Camera::getFov() const
{
	return C_fov;
}

float* const Xr_Camera::getAngle() const
{
	return (float* const) C_angle;
}

Xr_MATRIX_4X4* const Xr_Camera::getProjection() const
{
	return (Xr_MATRIX_4X4* const) C_projection;
}

Xr_FrustumBox* const Xr_Camera::getFrustumBox() const
{
	return (Xr_FrustumBox* const) C_box;
}
