#include <cmath>

#include "matrix.h"

#include "vector.h"

Xr_MATRIX_4X4::Xr_MATRIX_4X4() : M_values(0)
{
	int i;
	M_values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		M_values[i] = 0;
	}
}

Xr_MATRIX_4X4::Xr_MATRIX_4X4(float *values) : M_values(0)
{
	int i;
	M_values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		M_values[i] = values[i];
	}
}

Xr_MATRIX_4X4::Xr_MATRIX_4X4(Xr_MATRIX_4X4 const& matrix) : M_values(0)
{
	int i;
	float *values = matrix.getValues();
	M_values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		M_values[i] = values[i];
	}
}

Xr_MATRIX_4X4::~Xr_MATRIX_4X4()
{
	delete[] M_values;
}

float* const Xr_MATRIX_4X4::getValues() const
{
	return (float* const) M_values;
}

void Xr_MATRIX_4X4::loadIdentity()
{
	int i;
	for(i = 0; i < 16; i++)
	{
		M_values[i] = 0;
	}

	M_values[0] = 1;
	M_values[5] = 1;
	M_values[10] = 1;
	M_values[15] = 1;
}

void Xr_MATRIX_4X4::loadProjection(float fov, float aspect, float zNear, float zFar)
{
	int i;
	float *values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		values[i] = 0;
	}
	
	float f = (1 / tan((fov / 2) * M_PI / 180));
	
	values[0] = (f / aspect);
	values[5] = f;
	values[10] = (zFar + zNear) / (zNear - zFar);
	values[11] = (2 * zFar * zNear) / (zNear - zFar);
	values[14] = -1;
	
	Xr_MATRIX_4X4 projection(values);
	delete[] values;
	
	*this = *this * projection;
}

void Xr_MATRIX_4X4::lookAt(float eyeX, float eyeY, float eyeZ, float centerX, float centerY, float centerZ, float upX, float upY, float upZ)
{
	int i;
	float *values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		values[i] = 0;
	}
	
	Xr_Vector3D axe(upX, upY, upZ);
	Xr_Vector3D look(centerX - eyeX, centerY - eyeY, centerZ - eyeZ);
	Xr_Vector3D normal = look * axe;
	Xr_Vector3D axeRecalc = normal * look;
	
	normal.normalize();
	axeRecalc.normalize();
	look.normalize();
	
	values[0] = normal[0];
	values[1] = normal[1];
	values[2] = normal[2];
	values[4] = axeRecalc[0];
	values[5] = axeRecalc[1];
	values[6] = axeRecalc[2];
	values[8] = -look[0];
	values[9] = -look[1];
	values[10] = -look[2];
	values[15] = 1;
	
	Xr_MATRIX_4X4 lookAt(values);
	delete[] values;
	
	*this = *this * lookAt;
	translate(-eyeX, -eyeY, -eyeZ);
}

void Xr_MATRIX_4X4::translate(float x, float y, float z)
{
	int i;
	float *values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		values[i] = 0;
	}
	
	values[0] = 1;
	values[5] = 1;
	values[10] = 1;
	values[15] = 1;
	
	values[3] = x;
	values[7] = y;
	values[11] = z;
	
	Xr_MATRIX_4X4 translate(values);
	delete[] values;
	
	*this = *this * translate;
}

void Xr_MATRIX_4X4::rotate(float angle, float x, float y, float z)
{
	int i;
	float *values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		values[i] = 0;
	}
	
	float c = cos(angle);
	float s = sin(angle);
	Xr_Vector3D vec(x, y, z);
	
	vec.normalize();
	
	values[0] = (vec[0] * vec[0]) * (1 - c) + c;
	values[1] = (vec[0] * vec[1]) * (1 - c) - (vec[2] * s);
	values[2] = (vec[0] * vec[2]) * (1 - c) + (vec[1] * s);
	
	values[4] = (vec[1] * vec[0]) * (1 - c) + (vec[2] * s);
	values[5] = (vec[1] * vec[1]) * (1 - c) + c;
	values[6] = (vec[1] * vec[2]) * (1 - c) - (vec[0] * s);
	
	values[8] = (vec[0] * vec[2]) * (1 - c) - (vec[1] * s);
	values[9] = (vec[1] * vec[2]) * (1 - c) + (vec[0] * s);
	values[10] = (vec[2] * vec[2]) * (1 - c) + c;
	
	values[15] = 1;
	
	Xr_MATRIX_4X4 rotate(values);
	delete[] values;
	
	*this = *this * rotate;
}

void Xr_MATRIX_4X4::scale(float x, float y, float z)
{
	int i;
	float *values = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		values[i] = 0;
	}
	
	values[0] = x;
	values[5] = y;
	values[10] = z;
	values[15] = 1;
	
	Xr_MATRIX_4X4 scale(values);
	delete[] values;
	
	*this = *this * scale;
}

Xr_MATRIX_4X4& Xr_MATRIX_4X4::operator=(Xr_MATRIX_4X4 const& matrix)
{
	int i;
	float *values = matrix.getValues();

	for(i = 0; i < 16; i++)
	{
		M_values[i] = values[i];
	}

	return *this;
}

Xr_MATRIX_4X4 Xr_MATRIX_4X4::operator*(Xr_MATRIX_4X4 const& matrix)
{
	int i, j, k;
	float *result = new float[16];
	
	for(i = 0; i < 16; i++)
	{
		result[i] = 0;
	}
	
	float *values = matrix.getValues();

	for(k = 0; k < 4; k++)
	{
		for(j = 0; j < 4; j++)
		{
			for(i = 0; i < 4; i++)
			{
				result[(j * 4) + k] += M_values[(j * 4) + i] * values[(4 * i) + k];
			}
		}
	}

	Xr_MATRIX_4X4 res(result);
	delete[] result;
	return res;	
}
